@@ -29,9 +29,10 @@ use vars qw(@ISA @EXPORT);
 @EXPORT=qw(read_locale read_packages read_sources set_base
  get_missing_sources get_missing_binaries get_britney_list
  get_single_package get_archlist get_locale_roots get_base
- get_components get_britney_complaint get_missing_builddeps );
+ get_components get_britney_complaint get_missing_builddeps
+ get_suite_names get_unfrozen_suite_names );
 
-our $OURVERSION = "0.2.1";
+our $OURVERSION = "0.2.3";
 
 =head1 Synopsis
 
@@ -182,6 +183,67 @@ sub get_components
 	return \@cmpnts;
 }
 
+=head1 get_suite_names
+
+Parses the conf/distributions file for the specified repository and
+retrieves the list of configured suite names - a collated unique list
+of the Suite: and Codename: variables in conf/distributions.
+
+=cut
+
+sub get_suite_names
+{
+	my (@stanzas, $parser, $conf, $stanza);
+	my $repo = shift;
+	my %suites=();
+	return undef unless defined $base;
+	$parser = new Parse::DebControl;
+	my $input = "$base/$repo/conf/distributions";
+	$input =~ s://:/:g;
+	return undef unless (-f $input);
+	$conf = $parser->parse_file("$input", {'stripComments' => 'true'});
+	@stanzas = @$conf;
+	foreach $stanza (@stanzas)
+	{
+		$suites{$$stanza{'Codename'}}++;
+		$suites{$$stanza{'Suite'}}++;
+	}
+	my @ret = keys %suites;
+	return \@ret;
+}
+
+=head1 get_unfrozen_suite_names
+
+Parses the conf/distributions file for the specified repository and
+retrieves the list of configured suite names that are NOT ReadOnly.
+Codename: variables are NOT included.
+
+=cut
+
+sub get_unfrozen_suite_names
+{
+	my (@stanzas, $parser, $conf, $stanza);
+	my $repo = shift;
+	my %suites=();
+	return undef unless defined $base;
+	$parser = new Parse::DebControl;
+	my $input = "$base/$repo/conf/distributions";
+	$input =~ s://:/:g;
+	return undef unless (-f $input);
+	$conf = $parser->parse_file("$input", {'stripComments' => 'true'});
+	@stanzas = @$conf;
+	foreach $stanza (@stanzas)
+	{
+		if (exists $$stanza{'ReadOnly'})
+		{
+			next if ($$stanza{'ReadOnly'} eq 'Yes');
+		}
+		$suites{$$stanza{'Suite'}}++;
+	}
+	my @ret = keys %suites;
+	return \@ret;
+}
+
 =head1 set_base
 
 Set the base directory of the repository (repositories) to be
@@ -402,12 +464,21 @@ sub get_single_package
 		{
 			my $file = "$base/$repo/dists/$suite/$cmpnt/binary-${arch}/Packages";
 			$file =~ s://:/:g;
-			return undef if (not -f $file);
+			next if (not -f $file);
 			$fh = IO::File->new("$file") or die "$!\n";
 			$parser = Parse::Debian::Packages->new( $fh );
 			while (%package = $parser->next)
 			{
-				%list = (%package) if ($package{'Package'} eq $pkg);
+				if ($package{'Package'} eq $pkg)
+				{
+					if ($cmpnt eq 'main')
+					{
+						%list = (%package);
+						next;
+					}
+					my %detail = (%package);
+					$list{$cmpnt} = \%detail;
+				}
 			}
 			$fh->close;
 			undef $fh;
@@ -420,7 +491,7 @@ sub get_single_package
 	{
 		my $input = "$base/$repo/dists/$suite/$cmpnt/source/Sources.gz";
 		$input =~ s://:/:g;
-		return undef unless (-f $input);
+		next unless (-f $input);
 		my $z = new IO::Uncompress::Gunzip $input
 				or die "IO::Uncompress::Gunzip failed: $GunzipError\n";
 		$parser = Parse::Debian::Packages->new( $z );
@@ -638,6 +709,8 @@ sub get_britney_list
 		{
 			# check criterion 1.
 			next if (not defined $origtesting{$pkg}{$arch});
+			# skip source-only packages
+			next if (not defined $origunstable{$pkg}{$arch});
 			$skip++ if ($origunstable{$pkg}{$arch} ne $origtesting{$pkg}{$arch});
 		}
 		next if (defined $skip);
