@@ -4,6 +4,8 @@ use strict;
 use warnings;
 use IO::File;
 use Exporter;
+use POSIX qw(locale_h);
+use Locale::gettext;
 use File::Copy;
 use File::Basename;
 use Parse::Debian::Packages;
@@ -19,7 +21,8 @@ use vars qw/ %debianunstable %debiantesting %gripunstable %griptesting
  extend_filter grip_binary grip_source incoming_locale print_missing
  setup_repos update_filter set_repo_names set_noskip grip_britney
  print_testing_status update_repo set_dry_run migrate_missing
- check_syscall print_build_deps );
+ check_syscall print_build_deps scripts_version _g 
+ switch_component );
 
 =pod
 
@@ -277,24 +280,81 @@ C<Debian::Package::Compare::get_missing_builddeps> in a simple format
 sub print_build_deps
 {
 	$base = &get_base;
+	my $host = `dpkg-architecture -qDEB_BUILD_ARCH_CPU`;
+	chomp ($host);
 	my ($c, $s, $total);
 	my @lines=();
 	return undef unless (defined $filter_name and defined $grip_name);
 	my $missing = &get_missing_builddeps('unstable', "$filter_name", "$grip_name");
 	if (defined %$missing)
 	{
+		foreach my $p (sort keys %$missing)
+		{
+			my @k = split (/ /, $p);
+			my $pkg = $k[0];
+			next if (not defined $pkg);
+			my $res = `apt-cache show $pkg 2>/dev/null`;
+			chomp ($res);
+			next if ($res ne "");
+			my $cache = `apt-cache showpkg $pkg 2>/dev/null`;
+			$cache =~ s/\n/ /g;
+			if ($cache !~ /Reverse Provides:/)
+			{
+				print "INF: Unable to check '$pkg'. It may be an virtual ".
+				"alternative package or might not be available on '$host'\n";
+				next;
+			}
+			delete $$missing{$p};
+			$cache =~ /Reverse Provides: (.*)/;
+			my $match = $1;
+			$match =~ s/^ +//;
+			$match =~ s/ +$//;
+			my @pkgver = split (" ", $match);
+			my %pkgs = ();
+			foreach my $block (@pkgver)
+			{
+				next if ($block =~ /^[0-9]/);
+				next if ($block =~ /^$/);
+				$pkgs{$block}++;
+			}
+			my $rpkg = join (" | ", sort keys %pkgs);
+			next if ($rpkg eq "");
+			my @alternatives = ();
+			if (scalar keys %pkgs > 1)
+			{
+				@alternatives = sort keys %pkgs;
+				$rpkg = $alternatives[0];
+				my $end = scalar @alternatives;
+				if (scalar @alternatives > 2)
+				{
+					$end--;
+					print "INF: Alternatives for $rpkg: ";
+					print join (" | ", @alternatives[1..$end])."\n";
+				}
+				if (scalar @alternatives == 2)
+				{
+					$end--;
+					print "INF: Alternative for $rpkg: ";
+					print join (" ", $alternatives[$end])."\n";
+				}
+			}
+			my $check = `reprepro -b ${base}grip list sid $rpkg`;
+			chomp $check;
+			$check =~ s/.*: //g;
+			if ($check ne "")
+			{
+				print "INF: '$rpkg' Provides '$pkg' and already exists ".
+					"in the Grip repository, skipping '$rpkg'.\n";
+				next;
+			}
+			print "INF: Adding '$rpkg' to Provide: '$pkg'.\n";
+			$$missing{$rpkg} = \@alternatives;
+		}
 		$c = scalar (keys %$missing);
 		$s = ($c > 1) ? "build dependencies" : "build dependency";
-		print "INF: $c $s in the filter repository but not in Grip:\n";
+		print "INF: $c $s not yet in Grip:\n";
 		$c = 0;
 		print join (" ", sort keys %$missing)."\n";
-		print "INF: Checking the list . . . \n";
-		foreach my $pkg (sort keys %$missing)
-		{
-			my $check = `reprepro -b ${base}filter list sid $pkg`;
-			print "ERR: '$pkg' exists in filter:\n$check\n"
-				if ($check ne "");
-		}
 	}
 	print "INF: Done.\n";
 }
@@ -381,9 +441,7 @@ sub grip_britney
 	my ($c, $plural, $str, $vers);
 	return undef unless (defined $filter_name and defined $grip_name);
 	my $britney = &get_britney_list("$filter_name", "$grip_name");
-	# before trying the britney list, check for complaints.
 	my $complain = get_britney_complaint();
-	return $complain if (defined $complain);
 	$c = scalar (keys %$britney);
 	if ($c > 0)
 	{
@@ -414,7 +472,7 @@ sub grip_britney
 			(defined $dryrun) ? print "$str\n" : system ("$str 2>&1");
 		}
 	}
-	return undef;
+	return $complain;
 }
 
 =head1 migrate_missing
@@ -640,6 +698,7 @@ sub clean_incoming
 	$base = &get_base;
 	return undef unless (defined $grip_name);
 	print "INF: cleaning grip incoming directory\n";
+	print "INF: Stamp: " . `date`;
 	# stage removals in case the argument list gets too long.
 	# probably do an opendir readdir instead.
 	unlink <${base}${grip_name}/incoming/*.changes>;
@@ -647,6 +706,7 @@ sub clean_incoming
 	unlink <${base}${grip_name}/incoming/*.tdeb>;
 	unlink <${base}${grip_name}/incoming/*.dsc>;
 	unlink <${base}${grip_name}/incoming/*>;
+	system "rm -rf ${base}${grip_name}/incoming/*";
 }
 
 =head1 cleanup
@@ -803,13 +863,19 @@ sub incoming_locale
 		}
 		else
 		{
-			$str = "reprepro --ignore=extension -C $cmpnt -b ".
-			"${base}locale includedeb $suite ".
+			$str = "reprepro --ignore=extension -C $cmpnt ".
+			"--export=never -b ${base}locale includedeb $suite ".
 			"${base}${grip_name}/incoming/$t";
 			(defined $dryrun) ? print "$str\n" :
 				system ("$str 2>&1");
 		}
 	}
+	if (scalar @files > 1)
+	{
+		$str = "reprepro -v -b ${base}locale export $suite";
+		(defined $dryrun) ? print "$str\n" :
+			system ("$str 2>&1");
+	}
 }
 
 =head1 check_dirs
@@ -1084,19 +1150,10 @@ sub grip_source
 	chomp ($dir);
 	# need to get the .debs from the repo, emgrip will then grip these.
 	# filter only uses main
-	opendir (POOL, "$base/$filter_name/pool/main/$index/$package/")
-		or croak ("Cannot open $base/$filter_name/pool/main/$index/$package/ :$!");
 	# allow binNMUs etc. by adding .* to version
 	# this is source, so no $arch available.
-	my @debs=grep(/^.*${version}.*_.*\.deb$/, readdir (POOL));
-	closedir (POOL);
-	foreach my $f (@debs)
-	{
-		# filter only needs support for main
-		copy ("$base/$filter_name/pool/main/$index/$package/$f", $dir);
-	}
 	chdir ($dir);
-	print "INF: getting $dsc in $dir\n";
+	print "INF: getting source package: $dsc\n";
 	$str = "dget -ud file://${dsc}";
 	(defined $dryrun) ? print "$str\n" : system ($str);
 	$dsc = basename ($dsc);
@@ -1107,7 +1164,7 @@ sub grip_source
 	{
 		next if ($arch eq 'source');
 		$str = "dpkg-architecture -a${arch} -c \"DEB_BUILD_OPTIONS='usegrip'".
-			" /usr/bin/emgrip -o $base/$grip_name/incoming $dsc\"";
+			" /usr/bin/emgrip -o $base/$grip_name/incoming $dsc 2>&1 \" 2>/dev/null";
 		(defined $dryrun) ? print "$str\n" : system ($str);
 	}
 	chdir ("../");
@@ -1120,14 +1177,16 @@ sub grip_source
 	}
 	my $priority = (not exists $$single{'Priority'}) ?
 		"optional" : $$single{'Priority'};
-	print "INF: adding source package to grip using reprepro\n";
-	print "INF: Section: ".$$single{'Section'}."\n";
-	print "INF: Priority: ".$$single{'Priority'}."\n";
+	print "INF: adding source package to grip using reprepro. ";
+	print "Section: ".$$single{'Section'}." ";
+	print "Priority: ".$$single{'Priority'}."\n";
+	# all source packages go into main
 	$str = "reprepro -C main -V -S ".$$single{'Section'}." -P $priority".
 		" --ignore=brokensignatures -b ${base}${grip_name} includedsc $suite ".
 		"${base}${grip_name}/incoming/$dsc";
 	$retval = (defined $dryrun) ? print "$str\n" : system ("$str 2>&1");
 	&check_syscall ($retval);
+	# locale repo needs component to be main
 	print "INF: adding source package to locale using reprepro\n";
 	$str = "reprepro -C main -V -S ".$$single{'Section'}." -P ".$$single{'Priority'}.
 		" --ignore=brokensignatures -b ${base}locale includedsc $suite ".
@@ -1135,11 +1194,7 @@ sub grip_source
 	$retval = (defined $dryrun) ? print "$str\n" : system ("$str 2>&1");
 	&check_syscall ($retval);
 	# finally, add the gripped binaries
-	print "INF: adding binary packages to grip\n";
-	$str = "reprepro -V -b ${base}${grip_name} includedeb $suite ".
-		"${base}${grip_name}/incoming/*_${version}*.deb";
-	$retval = (defined $dryrun) ? print "$str\n" : system ("$str 2>&1");
-	&check_syscall ($retval);
+	# read list of components from grip repository
 	&incoming_locale ($a, $package, $version, $suite);
 	&clean_incoming();
 }
@@ -1153,72 +1208,79 @@ into locale.
 
 Once the binary package has been processed, runs F<&clean_incoming>.
 
+=head2 Components and Sections
+
+Each binary package belongs to a C<Section> that broadly describes
+the type of functionality in the package. If a component exists in
+the repository with the same name as the C<Section>, the package will
+be put into that component. Additionally, if the package name ends
+in C<-dev> or the C<Section> is devel, the C<dev> component is
+selected. If the package name ends in C<-doc> or the C<Section> is
+C<doc>, the C<doc> component is selected. If there is no matching
+component configured for the Grip repository, C<main> is used. Section
+names themselves are read from the filter repository - i.e. set by
+the Debian ftp-master overrides.
+
 =cut
 
 sub grip_binary
 {
 	my ($retval, $debbldopts, $outdir, $str);
 	my ($package, $version, $suite, $arch) = @_;
-	my @relocations = qw/ dev doc dbg /;
-	return undef unless (defined $filter_name and defined $grip_name);
+	if (not defined $filter_name or not defined $grip_name)
+	{
+		my $msg = "filter name or grip name not defined!";
+		warn "Unable to proceed: '$msg'\n";
+		return;
+	}
 	$base = &get_base;
 	# strip epochs
 	$version =~ s/^[0-9]://;
-	print "INF: Preparing to add single $package binary ($version)\n";
+	print "INF: Preparing to add single $package binary ($version) on $arch\n";
 	# identify the source package first.
 	# (use the requested suite in case of a package split)
 	# but handle -proposed-updates whereby the actual data is in
 	# stable but the built package goes into stable-proposed-updates.
 	my $proposed = $suite;
 	$proposed =~ s/-proposed-updates$//;
-	print "INF: get_single_package ($proposed, $grip_name, $package, $arch)\n";
 	my $single = get_single_package ($proposed, $grip_name, $package, $arch);
 	if (not exists $$single{'Section'})
 	{
 		$single = get_single_package ("unstable", "$filter_name", "$package", "$arch");
 	}
-	return if (not exists $$single{'Section'});
+	if (not exists $$single{'Section'})
+	{
+		my $msg = "get_single_package (unstable, $filter_name, $package, $arch)";
+		warn "Unable to get results for '$msg'\n";
+		return;
+	}
 	my $srcname = (defined $$single{'Source'}) ? $$single{'Source'} : $package;
 	$srcname =~ s/\(.*\)//;
 	chomp ($srcname);
 	$srcname =~ s/ //g;
-	print "INF: Using source package '$srcname'\n" if ($srcname ne $package);
-	print "INF: Section: ".$$single{'Section'}."\n";
-	print "INF: Priority: ".$$single{'Priority'}."\n";
-	# now find the .debs
-	my ($index) = convert_prefix ($srcname);
-	opendir (POOL, "${base}$filter_name/pool/main/$index/$srcname/")
-		or croak ("Unable to open ${base}$filter_name/pool/main/$index/$srcname/ : $!");
-	# copy across all debs for this version string for locale support
-	my @debs=grep(/^.*\Q${version}\E_${arch}|all\.deb$/, readdir (POOL));
-	closedir (POOL);
+	print "INF: Using source package '$srcname'. " if ($srcname ne $package);
+	print "Section: ".$single->{'Section'}." ";
+	print "Priority: ".$single->{'Priority'}."\n";
+	$single = get_single_package ($proposed, $filter_name, $package, $arch);
+	my $deb = $single->{'Filename'};
+	my @debs = ();
+	push @debs, "${base}$filter_name/$deb";
+	my $cmpnt = &switch_component ($$single{'Section'}, $package);
+	print "INF: Component: $cmpnt\n";
 	foreach my $d (@debs)
 	{
-		next unless (-f "${base}${filter_name}/pool/main/$index/$srcname/$d");
+		next unless (-f "$d");
 		$str = "dpkg-architecture -a${arch} -c \"DEB_BUILD_OPTIONS='usegrip' ".
-			"/usr/bin/emgrip -o ${base}${grip_name}/incoming ".
-			"${base}${filter_name}/pool/main/$index/$srcname/$d\"\n";
+			"/usr/bin/emgrip -o ${base}${grip_name}/incoming $d 2>&1 \" 2>/dev/null";
 		$retval = (defined $dryrun) ? print "$str\n" : system ($str);
 		&check_syscall($retval);
 		opendir (INCOMING, "${base}${grip_name}/incoming/")
 			or croak ("Cannot open ${base}${grip_name}/incoming/ :$!");
-		my @files=grep(/^\Q${package}_${version}\Eem._${arch}|all\.deb$/, readdir (INCOMING));
+		my @files=grep(/^\Q${package}_${version}\Eem._(${arch}|all)\.deb$/, readdir (INCOMING));
 		closedir (INCOMING);
 		foreach my $f (@files)
 		{
-			my $cmpnt = "-C main";
-			# handle relocations - reprepro will use main if fail.
-			foreach my $reloc (@relocations)
-			{
-				if ($f =~ /\-${reloc}_\Q${version}\E/)
-				{
-					my $actual = $reloc;
-					$actual =~ s/^dbg$/debug/;
-					print "INF: Moving $package to $actual section\n";
-					$cmpnt = "-C $actual";
-				}
-			}
-			$str = "reprepro -V $cmpnt -b ${base}${grip_name} includedeb".
+			$str = "reprepro -V -C $cmpnt -b ${base}${grip_name} includedeb".
 				" $suite ${base}${grip_name}/incoming/$f";
 			$retval = (defined $dryrun) ? print "$str\n" : system ("$str 2>&1");
 			&check_syscall ($retval);
@@ -1229,6 +1291,44 @@ sub grip_binary
 	}
 }
 
+=head1 switch_component
+
+Single routine to work out which components should be used for
+which sections.
+
+=cut
+
+sub switch_component {
+	my @relocations = qw/ dev doc dbg /;
+	my $cmpnt = "main";
+	my ($section, $pkg) = @_;
+	my $clist = get_components ('unstable', $grip_name);
+	# handle relocations - use main if fail.
+	$cmpnt = "dev"   if ($section eq 'libdevel');
+	$cmpnt = "dev"   if ($section eq 'devel');
+	$cmpnt = "debug" if ($section eq 'debug');
+	$cmpnt = "doc"   if ($section eq 'doc');
+	$cmpnt = "main"  if (not defined $clist);
+	$cmpnt = "main"  if (scalar (grep(/$cmpnt/, @$clist)) == 0);
+	print "INF: Switching component: $cmpnt\n"
+		if ($cmpnt ne "main");
+	foreach my $reloc (@relocations)
+	{
+		if ($pkg =~ /\-${reloc}$/)
+		{
+			my $actual = $reloc;
+			$actual =~ s/^dbg$/debug/;
+			$cmpnt = "$actual";
+		}
+	}
+	return $cmpnt;
+}
+
+sub scripts_version {
+	my $query = `dpkg-query -W -f='\${Version}' emdebian-grip-server`;
+	(defined $query) ? return $query : return "";
+}
+
 sub check_syscall
 {
 	my $signal = 0;
@@ -1237,12 +1337,16 @@ sub check_syscall
 	return if ($retval == 0);
 	$signal = $retval if ($retval < 256);
 	$retval /= 256 if ($retval >= 256);
-	carp ("INF: reprepro error: returned $retval")
+	print "INF: reprepro error: returned $retval\n" 
 		if ($retval != 0 and $signal == 0);
-	croak ("ERR: reprepro died after receiving signal $signal")
+	croak ("ERR: reprepro died after receiving signal $signal\n")
 		if ($signal > 0);
 }
 
+sub _g {
+	return gettext(shift);
+}
+
 =head1 AUTHOR
 
 Neil Williams, C<< <codehelp@debian.org> >>
